/*
 * main.c -- Main program for the GoAhead WebServer (DJGPP version)
 *
 * Copyright (c) GoAhead Software Inc., 1995-2000. All Rights Reserved.
 *
 * Modifications for Watt-32 by G. Vanem <giva@bgnett.no> Aug-1999
 *
 * See the file "license.txt" for usage and redistribution license requirements
 */

/******************************** Description *********************************/

/*
 *  Main program for for the GoAhead WebServer. This is a demonstration
 *  main program to initialize and configure the web server.
 */

/********************************* Includes ***********************************/

#include "../uemf.h"
#include "../wsIntrn.h"

#include <signal.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <setjmp.h>
#include <sys/time.h>
#include <tcp.h>

#ifdef WEBS_SSL_SUPPORT
  #include "../websSSL.h"
#endif

#ifdef USER_MANAGEMENT_SUPPORT
  #include "../um.h"
  #include "../emfdb.h"
  void formDefineUserMgmt (void);
#endif

#ifdef LOOPBACK_TEST
  #include "loopback.h"
#endif


/*
 * Change configuration here
 */
static char_t *rootWeb   = NULL;    /* Root web directory */
static char_t *password  = T("");   /* Security password */

static int     no_ssl     = 0;
static int     port       = 80;     /* Server port */
static int     retries    = 5;      /* Server port retries */
static int     traceLevel = 0;
static FILE   *traceFile  = stdout;

static jmp_buf sigint;

#ifdef LOOPBACK_TEST
static char_t *loop_url  = NULL;    /* loopback URI request */
static char_t *loop_post = NULL;    /* loopback post request */
static int     loop_test = 0;       /* do loopback test */
static int     loop_sock = -1;      /* loopback client socket */
#endif

static void sig_catch (int sig);

/************************** Forward Declarations **************************/

static int  initWebs (void);
static int  aspTest  (int eid, webs_t wp, int argc, char_t **argv);
static void formTest (webs_t wp, char_t *path, char_t *query);

#if B_STATS
static void printMemStats (int handle, char_t * fmt, ...);
static void memLeaks (void);
#endif

#ifdef USE_FORTIFY
static void ExitFortify (void)
{
  Fortify_ListAllMemory();
  Fortify_OutputStatistics();
}
#endif

static void usage (const char *argv0)
{
  printf ("Usage: %s [-dvs] [-r root] [-t level] [-l URL] [-p postdata]\n"
          "  options:  -d  debug mode\n"
          "            -v  print version and exit\n"
          "            -s  don't use SSL authentication\n"
          "            -r  specify web-root\n"
          "            -t  specify trace level\n"
          "            -l  request URL via loopback\n"
          "            -p  request post via loopback\n",
          argv0);
  exit (0);
}

/*
 * Main -- entry point for webs/djgpp
 */
int main (int argc, char **argv)
{
  int ch;

  while ((ch = getopt(argc, argv, "?vdr:st:l:p:")) != EOF)
     switch (ch)
     {
       case 'v': printf ("%s web-server. v%s, %s\n"
                         "%s\n", WEBS_NAME, WEBS_VERSION, __DATE__,
                         wattcpVersion());
                 return (0);
       case 'd': dbug_init();
                 break;
       case 'r': rootWeb = strdup (optarg);
                 break;
       case 's': no_ssl = 1;
                 break;
       case 't': traceLevel = atoi (optarg);
                 break;
       case 'l': loop_test = 1;
                 loop_url  = (char_t*)optarg;
                 break;
       case 'p': loop_test = 1;
                 loop_post = (char_t*)optarg;
                 break;
       case '?':
       default:  usage (argv[0]);
     }

#ifdef USE_FORTIFY
  Fortify_EnterScope();
  atexit (ExitFortify);
#endif

#ifndef NO_BALLOC
  /*
   * Initialize the memory allocator. Allow use of malloc and start 
   * with a 60K heap.  For each page request approx 8KB is allocated.
   * 60KB allows for several concurrent page requests.  If more space
   * is required, malloc will be used for the overflow.
   */
  bopen (NULL, (60*1024), B_USE_MALLOC);
#endif

  /* Initialize the web server
   */
  if (initWebs() < 0)
     return (-1);


#ifdef STATISTICS_SCREEN
  init_stats_screen();
#endif

#ifdef LOOPBACK_TEST
  if (loop_test)
  {
    if (loop_post)
         loop_sock = http_client (T("POST"), loop_post, NULL);
    else loop_sock = http_client (T("GET"), loop_url, NULL);

    if (loop_sock < 0)
       return (-1);
  }
#endif

  printf ("Press ^BREAK to quit\n");
  signal (SIGINT, sig_catch);
  setcbrk (0);
  if (setjmp (sigint))
  {
    printf ("SIGINT caught\n");
    goto quit;
  }

#ifdef WEBS_SSL_SUPPORT
  if (!no_ssl)
     websSSLOpen();
#endif

  /*
   * Basic event loop. socketReady() returns true when a socket is ready
   * for service. socketSelect() will block for max. 1 sec until an event
   * occurs on an allocated socket (-1 = poll all)
   * socketProcess() will actually do the servicing.
   */
  while (1)
  {
#if 0
    if (loop_sock >= 0 && socketReady(loop_sock))
#endif

    if (socketReady(-1) || socketSelect(-1, 1000))
       socketProcess (-1);

    websCgiCleanup();
    emfSchedProcess();  /* check timer handlers etc. */
  }

quit:

#ifdef WEBS_SSL_SUPPORT
  if (!no_ssl)
     websSSLClose();
#endif

#ifdef USER_MANAGEMENT_SUPPORT
  umClose();
#endif

 /* Close the socket module, report memory leaks and close
  * the memory allocator.
  */
  websCloseServer();
  socketClose();
#if B_STATS
  memLeaks();
#endif
  bclose();
  return (0);
}

/*
 * ^C/^Break signal handler
 */
static void sig_catch (int sig)
{
  signal (SIGINT, SIG_IGN);  /* ignore until application quits */
  longjmp (sigint, sig);
}

/*
 * Home page handler
 */
static int websHomePageHandler (webs_t wp, char_t *urlPrefix, char_t *webDir,
                                int arg, char_t *url, char_t *path, char_t *query)
{
  /* If the empty or "/" URL is invoked, redirect default URLs to
   * the home page
   */
  if (*url == '\0' || !gstrcmp(url, T("/")))
  {
    websRedirect (wp, T("default.html"));  /* was T("home.asp") */
    return (1);
  }
  return (0);
}

/*
 * Initialize the web server.
 */
static int initWebs (void)
{
  struct hostent *hp;
  struct in_addr intaddr;
  char   host[PATH_MAX], webdir[PATH_MAX], *cp;
  char_t wbuf[PATH_MAX];

  if (!rootWeb)  /* if not '-r' option used */
  {
    /* Set "$GOAHEAD_DIR" as the root web.
     */
    cp = getenv ("GOAHEAD_ROOT");
    if (cp)
         strncpy (webdir, cp, sizeof(webdir));
    else getcwd (webdir, sizeof(webdir));
    webdir [sizeof(webdir)-1] = '\0';
  }
  else
  {
    strncpy (webdir, rootWeb, sizeof(webdir)-1);
    webdir [sizeof(webdir)-1] = '\0';
  }
  
  for (cp = webdir; cp && *cp; )
  {
    cp = strchr (cp, '\\');
    if (cp)
       *cp++ = '/';
  }
  cp = strrchr (webdir, '/');
  if (cp && cp[1] == '\0')   /* remove trailing '/' */
    *cp = '\0';

  trace (0, T("web-page root-dir = `%s'\n"), webdir);

  /* Initialize the socket subsystem
   */
  if (socketOpen() < 0)
  {
    printf ("socketOpen() failed\n");
    return (-1);
  }

#ifdef USER_MANAGEMENT_SUPPORT

  /* Initialize the User Management database
   */
  umOpen();
  umRestore (T("umconfig.txt"));
#endif

  /* Define the local Ip address, host name, default home page and the
   * root web directory.
   */
  if (gethostname (host, sizeof(host)) < 0)
  {
    error (E_L, E_LOG, T("Can't get hostname"));
    return (-1);
  }
  if ((hp = gethostbyname (host)) == NULL)
  {
    error (E_L, E_LOG, T("Can't get host address"));
    return (-1);
  }
  memcpy ((char*)&intaddr, (char*)hp->h_addr_list[0], hp->h_length);

  /* Configure the web server options before opening the web server
   */
  websSetDefaultDir (webdir);

  cp = inet_ntoa(intaddr);
  ascToUni(wbuf, cp, min(strlen(cp) + 1, sizeof(wbuf)));
  websSetIpaddr(wbuf);
  ascToUni(wbuf, host, min(strlen(host) + 1, sizeof(wbuf)));
  websSetHost(wbuf);

  /* Configure the web server options before opening the web server
   */
#if 0
  websSetDefaultPage (T("default.asp"));
#else
  websSetDefaultPage (T("default.html"));
#endif
  websSetPassword (password);

  /* Open the web server on the given port. If that port is taken, try
   * the next sequential port for up to "retries" attempts.
   */
  websOpenServer (port, retries);

  /* First create the URL handlers. Note: handlers are called in sorted order
   * with the longest path handler examined first. Here we define the security
   * handler, forms handler and the default web page handler. All handlers
   * use same virtual root (webdir).
   */
  websUrlHandlerDefine (T(""),        NULL, 0, websSecurityHandler, WEBS_HANDLER_FIRST);
  websUrlHandlerDefine (T("/goform"), NULL, 0, websFormHandler,0);
  websUrlHandlerDefine (T("/cgi-bin"),NULL, 0, websCgiHandler, 0);
  websUrlHandlerDefine (T(""),        NULL, 0, websDefaultHandler, WEBS_HANDLER_LAST);

  /* Now define two test procedures. Replace these with your application
   * relevant ASP script procedures and form functions.
   */
  websAspDefine (T("aspTest"), aspTest);
  websFormDefine(T("formTest"),formTest);

  /* Create the Form handlers for the User Management pages
   */
#ifdef USER_MANAGEMENT_SUPPORT
  formDefineUserMgmt();
#endif

  /* Create a handler for the default home page
   */
  websUrlHandlerDefine (T("/"), NULL, 0, websHomePageHandler, 0);
  return (0);
}

/*
 * Test Javascript binding for ASP. This will be invoked when "aspTest" is
 * embedded in an ASP page. See web/asp.asp for usage. Set browser to
 * "localhost/asp.asp" to test.
 */
static int aspTest (int eid, webs_t wp, int argc, char_t **argv)
{
  char_t *name, *address;

  if (ejArgs (argc, argv, T("%s %s"), &name, &address) < 2)
  {
    websError (wp, 400, T("Insufficient args\n"));
    return (-1);
  }
  return websWrite (wp, T("Name: %s, Address %s"), name, address);
}

/*
 * Test form for posted data (in-memory CGI). This will be called when the
 * form in web/asp.asp is invoked. Set browser to "localhost/asp.asp" to test.
 */
static void formTest (webs_t wp, char_t * path, char_t * query)
{
  char_t *name    = websGetVar (wp, T("name"), T("Joe Smith"));
  char_t *address = websGetVar (wp, T("address"), T("1212 Milky Way Ave."));

  websHeader (wp);
  websWrite (wp, T("<body><h2>Name: %s, Address: %s</h2>\n"), name, address);
  websFooter (wp);
  websDone (wp, 200);
}


/*
 * Default error handler.  The developer should insert code to handle
 * error messages in the desired manner.
 */
void defaultErrorHandler (int etype, char_t * msg)
{
  if (msg)
     write (STDOUT_FILENO, msg, gstrlen(msg));
}

/*
 * Trace log. Customize this function to log trace output
 */
void defaultTraceHandler (int level, char_t *buf)
{
  if (buf && traceFile && traceLevel >= level)
     fputs (buf, traceFile);
}

/*
 * Returns a pointer to an allocated qualified unique temporary file name.
 * This filename must eventually be deleted with bfree();
 */
char_t *websGetCgiCommName (void)
{
  char_t *pname1 = tempnam (NULL, T("cgi"));
  char_t *pname2 = bstrdup (B_L, pname1);

  trace (8, T("websGetCgiCommName: `%s'\n"), pname1);
  free (pname1);
  return (pname2);
}

/*
 * Launch the CGI process and return a handle to it.
 */
int websLaunchCgiProc (char_t *cgiPath, char_t **argp, char_t **envp,
                       char_t *stdIn,   char_t  *stdOut)
{
  int fdin  = -1, hstdin  = -1;
  int fdout = -1, hstdout = -1;
  int rc    = 0;

  if (traceFile && traceLevel >= 9)
  {
    char_t *p;
    int     i;

    fprintf (traceFile, "path: `%s'\n", cgiPath);

    for (i = 0, p = *argp; p; i++, p = argp[i])
       fprintf (traceFile, "  arg[%d]: `%s'\n", i, p);
    for (i = 0, p = *envp; p; i++, p = envp[i])
       fprintf (traceFile, "  env[%d]: `%s'\n", i, p);
  }

  if ((fdin = open(stdIn, O_RDWR | O_CREAT, 0666)) < 0)
     goto DONE;
  if ((fdout = open(stdOut, O_RDWR | O_CREAT, 0666)) < 0)
     goto DONE;

  if ((hstdin = dup(STDIN_FILENO)) == -1 ||
      (hstdout = dup (STDOUT_FILENO)) == -1 ||
      dup2 (fdin, STDIN_FILENO) == -1 ||
      dup2 (fdout, STDOUT_FILENO) == -1)
    goto DONE;

  setmode (fdin, O_BINARY);
  setmode (fdout, O_BINARY);

  if (execve (cgiPath, argp, envp) == -1)
  {
    printf ("content-type: text/html\n\n"
            "Execution of cgi process failed\n");
    rc = -1;
  }

DONE:
  if (hstdout >= 0)
     dup2 (hstdout, 1);

  if (hstdin >= 0)
     dup2 (hstdin, 0);

  if (fdout >= 0)
     close (fdout);

  if (fdin >= 0)
     close (fdin);

  return (rc);
}

/*
 * Check the CGI process.  Return 0 if it does not exist; non 0 if it does.
 */
int websCheckCgiProc (int handle)
{
  return (0);
}

#if B_STATS
static void memLeaks (void)
{
  int fd = gopen (T("leak.txt"), O_CREAT|O_TRUNC);

  if (fd >= 0)
  {
    bstats (fd, printMemStats);
    close (fd);
  }
}

/*
 * Print memory usage / leaks
 */
static void printMemStats (int handle, char_t * fmt, ...)
{
  va_list args;
  char_t buf[256];

  va_start (args, fmt);
  vsprintf (buf, fmt, args);
  va_end (args);
  write (handle, buf, strlen (buf));
}
#endif


#ifdef NOT_USED
/*
 *    Structure to hold timer events
 */
typedef struct {
        void (*routine) (long arg);  /* Timer routine */
        long   arg;                  /* Argument to routine */
      } websTimer_t;

/* The following holds the pointer to an allocated websTimer_t structure .
 * Using this method only one timer can be active at a time, but
 * for the WebServer, this should be OK. 
 */
static websTimer_t *tp;

static void timerProc (int signo);

/*
 * Create a timer to invoke the routine in "delay" milliseconds.
 */
void *emfCreateTimer (int delay, void (*routine) (long arg), long arg)
{
  struct sigaction act;
  struct itimerval its = { {0, 0}, {0, 0} };

  if ((tp = balloc (B_L, sizeof (websTimer_t))))
  {
    tp->routine = routine;
    tp->arg = arg;
  }
  else
    return (NULL);

  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  act.sa_handler = timerProc;
  sigaction (SIGALRM, &act, NULL);

  /* convert delay millisecs to secs and usecs required by struct
   */
  its.it_value.tv_sec = delay / 1000;
  its.it_value.tv_usec = (delay % 1000) * 1000;

  if (setitimer (ITIMER_REAL, &its, NULL) == -1)
  {
    bfree (B_L, tp);
    return (NULL);
  }
  return (tp);
}

/*
 * Delete a timer
 */
void emfDeleteTimer (void *id)
{
  websTimer_t *wtp;
  struct itimerval its = { {0, 0}, {0, 0} };

  wtp = (websTimer_t *) id;
  setitimer (ITIMER_REAL, &its, NULL);
  bfree (B_L, wtp);
}

/*
 * Timer handler
 */
static void timerProc (int signo)
{
  websTimer_t wtp = *tp;

  /* Copy the timer structure to a local first and delete it before calling
   * the function, since the function could create another timer.  In this 
   * implementation, only one timer can be allocated at a time.
   */
  bfree (B_L, tp);
  (wtp.routine) (wtp.arg);
}
#endif


